﻿// v2.1

// ECMAScript 5 strict mode
"use strict";

assert2(cr, "cr namespace not created");
assert2(cr.plugins_, "cr.plugins_ not created");

/////////////////////////////////////
// Plugin class
cr.plugins_.BHT_TicTacToe_AI = function(runtime)
{
	this.runtime = runtime;
};

(function ()
{
    var CONST_PlayerX = 1;
    var CONST_PlayerO = 2;

	/////////////////////////////////////
	var pluginProto = cr.plugins_.BHT_TicTacToe_AI.prototype;
		
	/////////////////////////////////////
	// Object type class
	pluginProto.Type = function(plugin)
	{
		this.plugin = plugin;
		this.runtime = plugin.runtime;
	};

	var typeProto = pluginProto.Type.prototype;
	
    // All of the winning sets of three are stored here - across/down/diagonal.
	var winnerLines = new Array();

    // Board data.
	var CurrentBoard = new Array(9);

	// called on startup for each object type
	typeProto.onCreate = function()
	{
	};

	/////////////////////////////////////
	// Instance class
	pluginProto.Instance = function(type)
	{
		this.type = type;
		this.runtime = type.runtime;
	};
	
	var instanceProto = pluginProto.Instance.prototype;

	// called whenever an instance is created
	instanceProto.onCreate = function()
	{
		// note the object is sealed after this call; ensure any properties you'll ever need are set on the object
		
	    // Fill in the winning combinations
        // Wins across
		winnerLines.push(new Array(0,1,2));
		winnerLines.push(new Array(3,4,5));
		winnerLines.push(new Array(6,7,8));
		
        // Wins down
		winnerLines.push(new Array(0,3,6));
		winnerLines.push(new Array(1,4,7));
		winnerLines.push(new Array(2,5,8));
		
        // Wins diagonal
		winnerLines.push(new Array(0,4,8));
		winnerLines.push(new Array(2,4,6));
		
        // Clear our board
		for(var i=0; i<9; i++)
		{
			CurrentBoard[i] = 0;
		}
		
        // There is always a maximizing and a minimizing player. These identify which is which.
		this.maximizingPlayer = 0;
		this.minimizingPlayer = 0;
		
		// Since we need two values to get winner details, we cache the winner ID and row (in the winnerLines array, so we can mark the correct winning 'line').
		this.cachedWinnerID = -1;
		this.cachedWinnerRow = -1;
	};
	
	// only called if a layout object - draw to a canvas 2D context
	instanceProto.draw = function(ctx)
	{
	};
	
	// only called if a layout object in WebGL mode - draw to the WebGL context
	// 'glw' is not a WebGL context, it's a wrapper - you can find its methods in GLWrap.js in the install
	// directory or just copy what other plugins do.
	instanceProto.drawGL = function (glw)
	{
	};

	//////////////////////////////////////
	// Conditions
	pluginProto.cnds = {};
	var cnds = pluginProto.cnds;
	
	//////////////////////////////////////
	// Actions
	pluginProto.acts = {};
	var acts = pluginProto.acts;

	// Store a move for player X
	acts.SetMoveX = function (index)
	{
		this.cachedWinnerID = this.cachedWinnerIndex = -1;	// Reset the cached values as they are now invalid.
		CurrentBoard[index] = CONST_PlayerX;
	};

    // Store a move for player O
	acts.SetMoveO = function (index)
	{
		this.cachedWinnerID = this.cachedWinnerIndex = -1;	// Reset the cached values as they are now invalid.
		CurrentBoard[index] = CONST_PlayerO;
	};
	
	// Run through the winner-array and see if anyone won, and at which index they won.
	acts.CheckForWinner = function ()
	{
		this.cachedWinnerID = this.cachedWinnerIndex = -1;	// Reset the cached values as they are now invalid.
				
		for (var i=0; i<winnerLines.length; i++)
		{
			if(CurrentBoard[winnerLines[i][0]] == CONST_PlayerX && CurrentBoard[winnerLines[i][1]] == CONST_PlayerX && CurrentBoard[winnerLines[i][2]] == CONST_PlayerX)
			{
				this.cachedWinnerID = CONST_PlayerX;
				this.cachedWinnerRow = i;
				break;
			}
			else if(CurrentBoard[winnerLines[i][0]] == CONST_PlayerO && CurrentBoard[winnerLines[i][1]] == CONST_PlayerO && CurrentBoard[winnerLines[i][2]] == CONST_PlayerO)
			{
				this.cachedWinnerID = CONST_PlayerO;
				this.cachedWinnerRow = i;
				break;
			}
		}
		
		// Check if the game is finished (no more moves).
		if(this.cachedWinnerID == -1)
		{
			var cnt = 0;
			for(var i=0; i<9; i++)
			{
				if(CurrentBoard[i] != 0)
				{
					cnt++;
				}
			}
			// If board is full
			if(cnt == 9)
			{
				this.cachedWinnerID = 0;	// no-winner!
			}
		}
	}
	
	//////////////////////////////////////
	// Expressions
	pluginProto.exps = {};
	var exps = pluginProto.exps;
	
	// Perform the AI to get the next best move for player O.
	exps.GetMoveO = function(ret)
	{
		this.cachedWinnerID = this.cachedWinnerIndex = -1;	// Reset the cached values as they are now invalid.
		
		this.maximizingPlayer = CONST_PlayerO;
		this.minimizingPlayer = CONST_PlayerX;
		ret.set_int(this.RunMinimax(CurrentBoard, this.maximizingPlayer));
	}
	
	// Perform the AI to get the next best move for player X.
	exps.GetMoveX = function(ret)
	{
		this.cachedWinnerID = this.cachedWinnerIndex = -1;	// Reset the cached values as they are now invalid.
		
		this.maximizingPlayer = CONST_PlayerX;
		this.minimizingPlayer = CONST_PlayerO;
		ret.set_int(this.RunMinimax(CurrentBoard, this.maximizingPlayer));
	}

    // Perform the AI to get the next best move for player X.
	exps.GetWinner = function (ret)
	{
		ret.set_int(this.cachedWinnerID);
	}
	
    // Perform the AI to get the next best move for player O.
	exps.GetWinnerIndex = function (ret)
	{
		ret.set_int(this.cachedWinnerRow);
	}

	// Given a board, and a player, maximize the score for the Maximizing player.
	instanceProto.RunMinimax = function(board, currentPlayer)
	{
		// [0] is the score
		// [1] is the move
		var bestMove = this.Minimax(board, currentPlayer, 0);
		
		return bestMove[1];
	}
	
	// Given a current board, and current player, run the Mini-max algorithm.
	instanceProto.Minimax = function(board, currentPlayer, depth)
	{
		// Reset the score to 0, and the best-move to -1.
		var result = [];
		result[0] = 0;		// score
		result[1] = -1;		// best move
		
		var children = [];
		
		// Cycle through every possible new move, and build a new board with that move, for the current player.
		for(var i=0; i<9; i++)
		{
			// If this position is blank, add a move here.
			if(board[i] == 0)
			{
				// Copy board and add new move
				var childBoard = board.slice();	// copy
				childBoard[i] = currentPlayer;	// new move
				children.push(new BoardMove(childBoard, i));	// store on an array
			}
		}
		
		// if the current player is the Maximizing player, start the best-score at -100, so that any higher score will replace this value.
		// Otherwise the current player is the minimizing player so start the best-score at 100 so lower scores will replace this value.
		var bestScore = currentPlayer == this.maximizingPlayer ? -100 : 100;
		var currentScore = 0;
		var bestMove = -1;
		var currentResult = [];
		
		// Reached the bottom, so we have no score (for either player).
		if(children.length == 0)
		{
			bestScore = 0;
		}
		else
		{
			// Increase our depth counter;
			depth++;
			
			// Cycle through all of the possible board moves
			for(var n = 0; n< children.length; n++)
			{
				// If we are maximizing
				if(currentPlayer == this.maximizingPlayer)
				{
					// Is this board a winner for the ***maximizing*** player
					if(this.IsWinner(children[n].Board, this.maximizingPlayer))
					{
						currentScore = 10 - depth;	// Give a positive score.
					}
					else
					{
						// No, so run the algorithm for the ***minimizing*** player
						currentResult = this.Minimax(children[n].Board, this.minimizingPlayer, depth);
						currentScore = currentResult[0];	// Store the score
					}
					
					// If our new score is higher than the one we had
					if(currentScore > bestScore)
					{
						// Store the score and the move that got this score.
						bestScore = currentScore;
						bestMove = children[n].Move;
					}
				}
				else	// Current player is minimizing
				{
					// Is this board a winner for the ***minimizing*** player					
					if(this.IsWinner(children[n].Board, this.minimizingPlayer))
					{						
						currentScore = depth - 10;	// Give a negative score.
					}
					else
					{
						// No, so run the algorithm for the ***maximizing*** player
						currentResult = this.Minimax(children[n].Board, this.maximizingPlayer, depth);
						currentScore = currentResult[0];	// Store the score
					}
					
					// If our new score is lower than the one we had
					if(currentScore < bestScore)
					{
						// Store the score and the move that got this score.
						bestScore = currentScore;
						bestMove = children[n].Move;
					}
				}
			}
		}
		
		// Return our best-score and best-move.
		result[0] = bestScore;
		result[1] = bestMove;
	
		return result;
	}
	
	// Does this board produce a win for the given player
	instanceProto.IsWinner = function(board, player)
	{
		// Run through the winning-lines array to see if this player matches the positions of a winning combination.
		for (var i=0; i<winnerLines.length; i++)
		{
			if(board[winnerLines[i][0]] == player && board[winnerLines[i][1]] == player && board[winnerLines[i][2]] == player)
			{
				return true;
			}
		}
		return false;
	}
	
	// When we build a new possible board, we need to store the board and the move that was added. This just stores those two values for us, so we can place it in an array.
	function BoardMove(board, move)
	{
		this.Board = board;
		this.Move = move;	
	}

}());